#!/bin/sh
#
# Purpose:
#   Merges studies into one new study
#
#
# Synopsis:
#    merge.sh <descriptor file>
#
# Arguments:
#      descriptor file   : Contains information on studies to be merged
#
# Description:
#
# Merges studies listed in the descriptor file. The descriptor file shall contain
# fields per line in <fieldname>="<fieldvalue>" format. The following fields shall
# be used in the descriptor file
#
# | Studies       | space separated list of Study instance UIDs of the studies to be merged
# | PrimaryStudy  | Study Instance UID of the primary study
# | KeepReports   | true/false
#
# Note:
#
# If 'KeepReports" flag is "true", then reports from other studies are
# attached as addendums to the Primary Study's report
#
# A new Study Instance UID will be generated for the new study and for each object
# in the new study.
#
# The patient and study information gets copied from Primary study
#
# The new study will be moved to data/dicom and get registered.
#
# ENDDOC
#
#  Last Modified by : $Author$
#  Last Modified at : $Date$
#  RCS File Name    : $RCSfile$
#  RCS Revision     : $Revision$
#  Locked by        : $Locker$
#  Symbolic name    : $Name$
#  State            : $State$
#

. $HOME/etc/setenv.sh


LOGGER="/usr/bin/logger"
TAG="merge[$$]"

Date=`date +"%Y%m%d%H%M%S"`


# prints error message to standard error
# end exits with error code 1
#
# $1  - error message to be printed
#
error_message_()
{
	LOG_ERROR=`echo "$1" | sed 's;[^a-z_A-Z+0-9\.\-];_;g'`
	$LOGGER -p $LOG_OPER_FACILITY.error -t "$TAG" "ERROR STUDY mergestudy $LOG_USER - - - - - - - - - - - ${LOG_ERROR}"
	$LOG_HOME/logintodatabase.sh -p "ERROR" -l "STUDY" -a "mergestudy" -u "$LOG_USER" -m "${LOG_ERROR}" 
	echo $1 >&2
	cd $CWD

	exit 1
}


########################################################################
# Initializing
#


CWD=`pwd`

NEW_STUDY_DIR=$TMP/newstudy.$$

# recieve in environment variable
[ -z "$LOG_USER" ] && LOG_USER="-"

[ -z "$1" ] && error_message_ "Missing argument: Descriptor file"
[ ! -f "$1" ] && error_message_ "Descriptor file $1 does not exist"

. $1

[ -z "$Studies" -o -z "$PrimaryStudy" ] && error_message_ "Bad descriptor file $1"

# change order of Studies to Primary Study, Other Studies order
OrderedStudies="$PrimaryStudy"
for s in $Studies ; do
  [ "$s" != "$PrimaryStudy" ] && OrderedStudies="$OrderedStudies $s"
done
Studies="$OrderedStudies"


# Create new Study Instance UID
NEWSTUDY=`$UTILS_HOME/bin/createUID study`
[ -z "$NEWSTUDY" ] && error_message_ "Can not create new Study Instance UID"

NEWPATIENTNAME=""

REFERENCEID="mergestudies.$$"


#######################################################################
# Functions
#


# cleans up temporary changes
# and calls error_message_ if $1 is
# specified
#
# [$1]  - error message to be printed
#
cleanup_()
{

	# remove mergestudies.$$ references STUDYDB reference counter of the studies
	for study in $Studies ; do
 		$UTILS_HOME/bin/refcount -r $REFERENCEID -t STUDYDB -i "$study" >>$LOG/refcount.error 2>&1
 		[ $? -ne 0 ] && $LOGGER -p $LOG_FACILITY.error -t "$TAG" "Could not remove $REFERENCEID reference from STUDYDB/$study"
	done

	[ -d "${NEW_STUDY_DIR}" ] && rm -fr ${NEW_STUDY_DIR}

	if [ -n "$1" ] ; then
		# locks should only be released in case of errors
		# in case of success locks will be released in obsolete tasks
		if [ -n "${Studies}" ] ; then
			for s in ${Studies} ; do
				rm ${LOCK_VAR}/r_${s} 2>/dev/null
				# also remove moveto. files - may cause problems for late objects or reindex
				rm ${TMP}/moveto.${s} 2>/dev/null
			done
		fi

		error_message_ "$1"
	fi

	cd $CWD

	return 0
}


#
# dump patient and study information from DICOM file
#
# $1 - file to be dumped
# $2 - output file
#
dumpfile_()
{
	$DCMTK_HOME/bin/dcmdump +C +L +P "0032,000c" +P "0010,0010" +P "0010,0020" +P "0010,0030" \
			+P "0010,0032" +P "0010,0040" +P "0008,0020" +P "0008,0030" +P "0008,0090" \
			+P "0020,0010" +P "0008,0050" +P "0008,0080" +P "0008,1030" +P "0008,1048" \
			+P "0008,1060" +P "0008,0061" +P "0008,1050" +P "0008,1010" +P "0018,0015" \
			+P "0040,0241" "$1" > "$2"
	[ $? -ne 0 ] && cleanup_ "Error dumping $1"
}

# will be used to detect whether an SRt file is a 3rd party or not
eRADCreatorUID="1.2.826.0.1.3680043.2.93.0.99"

#
# Detects whether an object was created by eRAD
# or is a 3rd party
#
# Checks the value of the (0008,0014) Instance Creator UID attribute
#
# $1 - path to the object to check
#
# return 0 - was created by eRAD 
#        1 - 3rd party object
#        2 - an error occurred 
#
function isOurSR_() {
	local file="$1"
	local creatorLine
	local creatorUID

	creatorLine=$( ${DCMTK_HOME}/bin/dcmdump +P "0008,0014" "${file}" )
	[ $? -ne 0 ] && return 2
	[ -z "${creatorLine}" ] && return 1

	creatorUID=$( sed 's;^.*\[\(.*\)\].*$;\1;' <<< "${creatorLine}" )

	[ "${creatorUID}" = "${eRADCreatorUID}" ] && return 0
	return 1
}




#########################################################################################
#
# Main
#
#########################################################################################

# add mergestudies.$$ reference STUDYDB reference counter of the studies to make sure the
# studies will not be deleted before the merge completes
for study in $Studies ; do
	$UTILS_HOME/bin/refcount -E -a $REFERENCEID -t STUDYDB -i "$study" >>$LOG/refcount.error 2>&1
	[ $? -ne 0 ] && cleanup_ "Could not add reference to study $study"
	echo "NewStudyUID=\"$NEWSTUDY\"" > ${TMP}/moveto.${study}
done


cd $TMP

minsopiname=`echo \
	"select FNAME from Dcobject, Dcstudy where Dcstudy.STYIUID='$PrimaryStudy' and Dcobject.STYIUID='$PrimaryStudy' and FOLDER='' and SOPIUID=MINSOPIUID;" | \
	$MYSQL_HOME/bin/mysql -N $DBNAME 2>/dev/null`
[ $? -ne 0 ] && cleanup_ "MySQL Error"


pbrname=`echo \
	"select FNAME from Dcobject where STYIUID='$PrimaryStudy' and SOPCLUID='$PBREPORT_SOPCLASS';" | \
	$MYSQL_HOME/bin/mysql -N $DBNAME 2>/dev/null`
[ $? -ne 0 ] && cleanup_ "MySQL Error"

PrimaryStudyDir=`$REPOSITORY_HANDLER_HOME/bin/locateResource $REPOSITORY_DATA_DICOM $PrimaryStudy`
[ $? -ne 0 ] && cleanup_ "could not locate primary study in repository"

minsopifile="$PrimaryStudyDir/$minsopiname"

pbrfile="$PrimaryStudyDir/$pbrname"

[ ! -f "$minsopifile" ] && cleanup_ "$minsopifile does not exist"

mkdir ${NEW_STUDY_DIR}
[ $? -ne 0 ] && cleanup_ "Can not create temporary directory for new study"

mkdir ${NEW_STUDY_DIR}/src
[ $? -ne 0 ] && cleanup_ "Can not create temporary directory for new study"

mkdir ${NEW_STUDY_DIR}/dest
[ $? -ne 0 ] && cleanup_ "Can not create temporary directory for new study"

MODFILE="${NEW_STUDY_DIR}/modfile"
FILELIST="${NEW_STUDY_DIR}/filelist"
UIDLIST="${NEW_STUDY_DIR}/uidlist"

#
# MODFILE will contain patient and study data to be copied into new study's objects
#

dumpfile_ $minsopifile $MODFILE.dmp

$DCMTK_HOME/bin/dump2dcm +C $MODFILE.dmp $MODFILE
[ $? -ne 0 ] && cleanup_ "Error creating DICOM file from dump: $MODFILE"


if [ -f "$pbrfile" ] ; then
	dumpfile_ $pbrfile ${NEW_STUDY_DIR}/pbrdump

	# override attributes values of minsopifile with values from pbrfile
	mod="${NEW_STUDY_DIR}/pbrmodfile"

	echo "(f215,1043) SQ # PBOverWriteSequence" > ${mod}
	echo "(fffe,e000) na # item" >> ${mod}
	cat ${NEW_STUDY_DIR}/pbrdump >>  ${mod}
	echo "(fffe,e00d) na # ItemDelimitationItem" >> ${mod}
	echo "(fffe,e0dd) na # SequenceDelimitationItem" >> ${mod}

	$DCMTK_HOME/bin/dump2dcm +C ${mod} ${mod}.dcm
	[ $? -ne 0 ] && cleanup_ "Error creating DICOM file from dump: ${mod}.dcm"

	$DCVIEWER_HOME/dcmmod -f ${MODFILE} -m ${mod}.dcm
	[ $? -ne 0 ] && cleanup_ "Error editing ${studymodfile}"

	rm ${NEW_STUDY_DIR}/pbrdump
	rm ${mod}
	rm ${mod}.dcm
fi


# Create list of images to be copied

for study in $Studies ; do
	studydir=`$REPOSITORY_HANDLER_HOME/bin/locateResource $REPOSITORY_DATA_DICOM $study`
	[ $? -ne 0 ] && cleanup_ "Study $study directory could not be located in repository"

	if [ -f "$CASES_HOME/studyinfo.shi" ] ; then
		. $CASES_HOME/studyinfo.shi
	fi

	[ -z "$DB_PNAME" ] && DB_PNAME="-"
	[ -z "$DB_MODALITY" ] && DB_MODALITY="-"

	[ -z "$NEWPATIENTNAME" ] && NEWPATIENTNAME=${DB_PNAME}
	$LOGGER -p $LOG_OPER_FACILITY.info -t "$TAG" "INFO STUDY mergestudy $LOG_USER - $study $DB_PNAME $DB_MODALITY - - - - - - - into_${NEWSTUDY}_${NEWPATIENTNAME}"
	$LOG_HOME/logintodatabase.sh -p "INFO" -l "STUDY" -a "mergestudy" -u "$LOG_USER" -s "$study" -p "$DB_PNAME" -m "into_${NEWSTUDY}_${NEWPATIENTNAME}" 

	for file in `ls -v $studydir/` ; do
		[ "$file" = "storestate.rec" ] && continue
		[ "${file#PbR.}" != "$file" -a "$study" != "$PrimaryStudy" ] && continue
		[ "${file#PbD.}" != "$file" ] && continue
		[ "${file#SRt}" != "$file" ] && continue

		cp $studydir/$file ${NEW_STUDY_DIR}/src/
		echo "${NEW_STUDY_DIR}/src/$file" >> $FILELIST

	done
done


# Copy SRt Reports

SRCONVERT=$TMP/srconvert.$NEWSTUDY
rm $SRCONVERT 2>/dev/null

PBDLIST=""

LATEST_SRT=""
MAX_PREDEC_NUM="-1"

SHORTCUTS=""

# Primary Study is the first in Studies list
for study in $Studies ; do
	studydir=`$REPOSITORY_HANDLER_HOME/bin/locateResource $REPOSITORY_DATA_DICOM $study`
	[ $? -ne 0 ] && cleanup_ "Study $study directory could not be located in repository"

	shortcuts=`echo "select STYIUID from $STUDYTBL where SRCSTUDY='${study}' AND STYIUID<>'${study}';" | ${MYSQL_HOME}/bin/mysql -N $DBNAME`
	[ -n "$shortcuts" ] && SHORTCUTS="$SHORTCUTS $shortcuts"

	# For copying non report objects
	echo "Replace $study by $NEWSTUDY" >> $UIDLIST

	# Only copy reports if the study is the Primary study or the KeepReports flag is turned on
	if [ "$study" != "$PrimaryStudy" -a "$KeepReports" != "true" ] ; then
		$LOGGER -p $LOG_FACILITY.debug -t "$TAG" "Do not copy reports: [$study != $PrimaryStudy] AND [$KeepReports != true]"
		continue
	else
		$LOGGER -p $LOG_FACILITY.debug -t "$TAG" "Copy reports: [$study = $PrimaryStudy] OR [$KeepReports = true]"
	fi

#	echo "Copy SR for $study [Latest SRT: $LATEST_SRT]"

	if [ -z "$LATEST_SRT" ] ; then
		# First study with report
		# Simply copy the SRt files

		for srfile in `ls $studydir/ | grep ^SRt` ; do
			cp $studydir/$srfile ${NEW_STUDY_DIR}/src/
			echo "${NEW_STUDY_DIR}/src/$srfile" >> $FILELIST

			if isOurSR_ "$studydir/$srfile" ; then
				# Set the SRt as Latest SRt only if it was created by eRAD (leave 3rd party SRt files as is) 

				PREDEC_NUM=`$DCMTK_HOME/bin/dcmdump +P "0008,1155" "${NEW_STUDY_DIR}/src/$srfile" | wc -l`
	
				if [ $PREDEC_NUM -gt $MAX_PREDEC_NUM ] ; then
					LATEST_SRT="$srfile"
					MAX_PREDEC_NUM="$PREDEC_NUM"
				fi
			fi
		done
	else
		# sort SR files of study by number of predecessor document entries
		# and add predecessor documents from latest sr

		SRLISTFILE="$NEW_STUDY_DIR/modifySRt.$study"

		for srfile in `ls $studydir/ | grep ^SRt` ; do
			cp $studydir/$srfile ${NEW_STUDY_DIR}/src/
			echo "${NEW_STUDY_DIR}/src/$srfile" >> $FILELIST

			if isOurSR_ "$studydir/$srfile" ; then
				# Add SRt file to the eRAD report chain (predecessor) only it was created by eRAD
				# (leave 3rd party SRt files as is)
 
				PREDEC_NUM=`$DCMTK_HOME/bin/dcmdump +P "0008,1155"  "${NEW_STUDY_DIR}/src/$srfile" | wc -l`
				echo "$PREDEC_NUM $srfile" >> $SRLISTFILE
			fi
		done

		if [ -f "$SRLISTFILE" ] ; then
			for srfile in `sort -n $SRLISTFILE | awk '{print $2}'` ; do
				echo "addSR \"${NEW_STUDY_DIR}/src/$LATEST_SRT\" \"${NEW_STUDY_DIR}/src/$srfile"

				$SR_HOME/bin/addSR "${NEW_STUDY_DIR}/src/$LATEST_SRT" "${NEW_STUDY_DIR}/src/$srfile"
				[ $? -ne 0 ] && cleanup_ "Error adding SRt files of $study"
				LATEST_SRT="$srfile"
			done

			rm -f $SRLISTFILE
		fi
	fi

	for pbd in `ls $studydir/ | grep ^PbD 2>/dev/null` ; do
		cp $studydir/$pbd ${NEW_STUDY_DIR}/src/
		PBDLIST="$PBDLIST $pbd"
	done

done

# echo "PBD LIST: $PBDLIST"

if [ -n "$PBDLIST" ] ; then
	NEWPBD=""
	NEWDLIST=""

	for pbd in $PBDLIST ; do
		if [ -z "$NEWPBD" ] ; then
			OBJNAME=`basename $pbd`
			DIRNAME=`dirname $pbd`
#			study=`basename $DIRNAME`

			NEWPBD="${NEW_STUDY_DIR}/src/$pbd"
			echo "${NEW_STUDY_DIR}/src/$pbd" >> $FILELIST
		else
			NEWDLIST="$NEWDLIST ${NEW_STUDY_DIR}/src/$pbd"
		fi
	done

	if [ -n "$NEWDLIST" ] ; then
		echo $SR_HOME/bin/mergedictations $NEWPBD $NEWDLIST
		$SR_HOME/bin/mergedictations $NEWPBD $NEWDLIST
		[ $? -ne 0 ] && cleanup_ "Error merging Dictation files"
	fi

fi


# echo "$UTILS_HOME/bin/copyDicomFiles -d ${NEW_STUDY_DIR}/dest -f $FILELIST -m ${MODFILE} -l $UIDLIST"

# copy files from src to dest
#	$UTILS_HOME/bin/copyDicomFiles -d ${NEW_STUDY_DIR}/dest -f $FILELIST -m ${MODFILE} -l $UIDLIST >${NEW_STUDY_DIR}/out
$UTILS_HOME/bin/copyDicomFiles -d ${NEW_STUDY_DIR}/dest -f $FILELIST -m ${MODFILE} -l $UIDLIST > ${TMP}/moveto.${NEWSTUDY}.uids 2>/dev/null
if [ $? -ne 0 ]; then
	cleanup_ "Copying DICOM objects failed"
	exit 1
fi

cp ${MODFILE} ${TMP}/moveto.${NEWSTUDY}.mod


# Create .info/storestate.rec file with StagedDevices and ForwardDevices copied from originals
StoreStateArgs="--study=\"${NEWSTUDY}\" --processmode=\"Process\" --owner=\"yes\" --derived=\"merge\""
for study in $Studies ; do
	StagedDevices=""
	ForwardDevices=""

	. $CASES_HOME/includeStorestate.shi "$study"
	for dev in $StagedDevices ; do
		StoreStateArgs="$StoreStateArgs --setstaged=\"$dev\""
	done

	for dev in $ForwardDevices ; do
		StoreStateArgs="$StoreStateArgs --fwddevice=\"$dev\""
	done
done

eval ${UTILS_HOME}/bin/storestate $StoreStateArgs

# Create derivation info file in .info of the new study
newStudyDir=`$REPOSITORY_HANDLER_HOME/bin/locateResource $REPOSITORY_DATA_DICOM ${NEWSTUDY}`
if [ $? -eq 0 -a -n "${newStudyDir}" ] ; then
	derivationFile="${newStudyDir}/.info/derivationInfo"
	touch "${derivationFile}"
	for study in $Studies ; do
		echo "${study} merge ${NEWSTUDY}" >> "${derivationFile}"
	done
fi

# Move new objects to IMAGE_TMP and start registering them
TaskUser="$LOG_USER"
[ "-" = "$TaskUser" ] && TaskUser=""

# make sure PbR is registered first
for file in `ls ${NEW_STUDY_DIR}/dest/ | grep ^PbR` ; do
	tempFile="$IMAGE_TMP/${NEWSTUDY}_${file}_$$"
	mv "${NEW_STUDY_DIR}/dest/$file" "$tempFile"

	$UTILS_HOME/bin/storeDICOMFile -f "${file}" -s "${NEWSTUDY}" -t "$tempFile" -D "DELTA_STORESCP_REG_MERGE_PBR" -U "${TaskUser}" -N "merging studies"
	if [ $? -ne 0 ] ; then
		# placing the StorescpReg task failed
		rm "$tempFile" 2>/dev/null
	fi
done

# register remaining (non PbR) objects 
for file in `ls ${NEW_STUDY_DIR}/dest/ | grep -v ^PbR;` ; do
	tempFile="$IMAGE_TMP/${NEWSTUDY}_${file}_$$"
	mv "${NEW_STUDY_DIR}/dest/$file" "$tempFile"

	$UTILS_HOME/bin/storeDICOMFile -f "${file}" -s "${NEWSTUDY}" -t "$tempFile" -D "DELTA_STORESCP_REG_MERGE" -U "${TaskUser}" -N "merging studies"
	if [ $? -ne 0 ] ; then
		# placing the StorescpReg task failed
		rm "$tempFile" 2>/dev/null
	fi
done


rmdir ${NEW_STUDY_DIR}/dest 2>&1

# Create shortcuts for the new study if needed
RDATE=`date +"%Y%m%d%H%M%S"`
for shortcut in $SHORTCUTS ; do
	$LOGGER -p $LOG_FACILITY.debug -t "$TAG" "${SR_HOME}/bin/createLink -M $shortcut -s $NEWSTUDY > tmp/out.$$.$NEWSTUDY 2> tmp/err.$$.$NEWSTUDY"
	${SR_HOME}/bin/createLink -u "${LOG_USER}" -M $shortcut -s $NEWSTUDY > "$TMP/out.$$.$NEWSTUDY" 2> "$TMP/err.$$.$NEWSTUDY"
	
	# remove old shortcut
	$TASKD_HOME/createtaskfileapp -p "RM_STUDY" -t "class" -c "${CASES_HOME#$HOME/component/}.RmStudy -s \"$shortcut\""
done

#check if all file was copied to NEW_STUDY_DIR
#call moveLateObject.sh for late ones
for study in $Studies ; do
	studydir=`$REPOSITORY_HANDLER_HOME/bin/locateResource $REPOSITORY_DATA_DICOM $study`
	[ $? -ne 0 ] && cleanup_ "Study $study directory could not be located in repository"

	for file in `ls -v $studydir/` ; do
		[ "$file" = "storestate.rec" ] && continue
		[ "${file#PbR.}" != "$file" -a "$study" != "$PrimaryStudy" ] && continue
		[ "${file#PbD.}" != "$file" ] && continue
		[ "${file#SRt}" != "$file" ] && continue
		[ -e "${NEW_STUDY_DIR}/src/${file}" ] && continue

		if [ ! -d "${TMP}/obsolete/${study}" ] ; then
			mkdir -p ${TMP}/obsolete/${study}
		fi

		mv $studydir/$file ${TMP}/obsolete/${study}/${file}
		
		$TASKD_HOME/createtaskfileapp -p "MOVE_LATE_OBJECT" -t "sh" -c "$CASES_HOME/moveLateObjects.sh $file $study"
	done
done



#obsolete original studies
RDATE=`date +"%Y%m%d%H%M%S"`
for study in $Studies ; do
	echo "UIDFile=\"${TMP}/moveto.${NEWSTUDY}.uids\"" >> ${TMP}/moveto.${study}
	echo "ModFile=\"${TMP}/moveto.${NEWSTUDY}.mod\"" >> ${TMP}/moveto.${study}
	
	$TASKD_HOME/createtaskfileapp -p "OBSOLETE_STUDY" -t "class" -c "${CASES_HOME#$HOME/component/}.ObsoleteStudy -s $study"
done

$LOGGER -p $LOG_OPER_FACILITY.info -t "$TAG" "INFO STUDY mergestudy $LOG_USER - ${NEWSTUDY} ${NEWPATIENTNAME} - - - - - - - - study_created"
$LOG_HOME/logintodatabase.sh -p "INFO" -l "STUDY" -a "mergestudy" -u "$LOG_USER" -s "${NEWSTUDY}" -p "${NEWPATIENTNAME}" -m "study_created" 

cleanup_
